package net.garstangs.layout.relative;

import java.util.LinkedHashMap;
import android.content.Context;
import android.database.Cursor;
import android.graphics.Bitmap;
import android.graphics.Rect;
import android.graphics.Typeface;
import android.text.InputFilter;
import android.text.InputType;
import android.text.TextUtils;
import android.text.method.NumberKeyListener;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnTouchListener;
import android.view.WindowManager;
import android.view.inputmethod.EditorInfo;
import android.widget.Button;
import android.widget.EditText;
import android.widget.RelativeLayout;
import android.widget.RelativeLayout.LayoutParams;
import android.widget.Spinner;
import android.widget.ScrollView;
import android.widget.TextView;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.ActivityObject;
import anywheresoftware.b4a.BA.Events;
import anywheresoftware.b4a.BA.Hide;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.BALayout;
import anywheresoftware.b4a.objects.AutoCompleteEditTextWrapper;
import anywheresoftware.b4a.objects.ButtonWrapper;
import anywheresoftware.b4a.objects.LabelWrapper;
import anywheresoftware.b4a.objects.SpinnerWrapper;
import anywheresoftware.b4a.objects.ViewWrapper;
import anywheresoftware.b4a.objects.collections.List;
import anywheresoftware.b4a.objects.collections.Map;
import anywheresoftware.b4a.objects.collections.Map.MyMap;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.Permissions;

@Permissions(values = {}) // Permissions must always be declared otherwise your app or library wont work.
@Events(values={"imeHeightChanged (NewHeight As Int, OldHeight As Int)", "imeAction As Boolean"})
@Author("Roger Garstang")

@ActivityObject
@ShortName("RelativeLayout")
@Version(1.00f)
public class MyRelativeLayoutWrapper extends ViewWrapper<ScrollView> {
	
	/**
     * Aligns the View with the Parent's Top [Set yView to Null in AddView()].
     */
	public static int getyAlign_ParentTop() { return RelativeLayout.ALIGN_PARENT_TOP;} // 10
	/**
     * Aligns the View with the Parent's Vertical Center [Set yView to Null in AddView()].
     */
	public static int getyAlign_ParentCentVert() { return RelativeLayout.CENTER_VERTICAL;} // 15
	/**
     * Aligns the View with another View's Top [Set yView to the Referenced View in AddView()].
     */
	public static int getyAlign_ViewTop() { return RelativeLayout.ALIGN_TOP;} // 6
	/**
     * Aligns the View with another View's Bottom [Set yView to the Referenced View in AddView()].
     */
	public static int getyAlign_ViewBottom() { return RelativeLayout.ALIGN_BOTTOM;} // 8
	/**
     * Aligns the View with another View's Text Baseline [Set yView to the Referenced View in AddView()].
     */
	public static int getyAlign_ViewBaseline() { return RelativeLayout.ALIGN_BASELINE;} // 4
	/**
     * Aligns the View Above another View [Set yView to the Referenced View in AddView()].
     */
	public static int getyAlign_AboveView() { return RelativeLayout.ABOVE;} // 2
	/**
     * Aligns the View Below another View [Set yView to the Referenced View in AddView()].
     */
	public static int getyAlign_BelowView() { return RelativeLayout.BELOW;} // 3
	/**
     * Aligns the View with the Parent's Left [Set xView to Null in AddView()].
     */
	public static int getxAlign_ParentLeft() { return RelativeLayout.ALIGN_PARENT_LEFT;} // 9
	/**
     * Aligns the View with the Parent's Right [Set xView to Null in AddView()].
     */
	public static int getxAlign_ParentRight() { return RelativeLayout.ALIGN_PARENT_RIGHT;} // 11
	/**
     * Aligns the View with the Parent's Horizontal Center [Set xView to Null in AddView()].
     */
	public static int getxAlign_ParentCentHorz() { return RelativeLayout.CENTER_HORIZONTAL;} // 14
	/**
     * Aligns the View with another View's Left [Set xView to the Referenced View in AddView()].
     */
	public static int getxAlign_ViewLeft() { return RelativeLayout.ALIGN_LEFT;} // 5
	/**
     * Aligns the View with another View's Right [Set xView to the Referenced View in AddView()].
     */
	public static int getxAlign_ViewRight() { return RelativeLayout.ALIGN_RIGHT;} // 7
	/**
     * Aligns the View to the Left of another View [Set xView to the Referenced View in AddView()].
     */
	public static int getxAlign_LeftOfView() { return RelativeLayout.LEFT_OF;} // 0
	/**
     * Aligns the View to the Right of another View [Set xView to the Referenced View in AddView()].
     */
	public static int getxAlign_RightOfView() { return RelativeLayout.RIGHT_OF;} // 1
	/**
     * Expands the View to the size needed for the content it contains.
     */
	public static int getSize_WrapContent() { return LayoutParams.WRAP_CONTENT;}
	/**
     * Expands the View to Fill the Parent.
     */
	public static int getSize_MatchParent() { return LayoutParams.MATCH_PARENT;}
	/**
     * No Action Button.
     */
	public static int getActionBtn_None() { return EditorInfo.IME_ACTION_NONE;} // 1
	/**
     * Shows Go Button and Fires ViewEventPrefix_imeAction Event.
     */
	public static int getActionBtn_Go() { return EditorInfo.IME_ACTION_GO;} // 2
	/**
     * Shows Search Button and Fires ViewEventPrefix_imeAction Event.
     */
	public static int getActionBtn_Search() { return EditorInfo.IME_ACTION_SEARCH;} // 3
	/**
     * Shows Send Button and Fires ViewEventPrefix_imeAction Event.
     */
	public static int getActionBtn_Send() { return EditorInfo.IME_ACTION_SEND;} // 4
	/**
     * Shows Next Button and Navigates to Next View.
     */
	public static int getActionBtn_Next() { return EditorInfo.IME_ACTION_NEXT;} // 5
	/**
     * Shows Done Button and Fires ViewEventPrefix_imeAction Event.
     */
	public static int getActionBtn_Done() { return EditorInfo.IME_ACTION_DONE;} // 6
	/**
     * Shows Previous Button and Navigates to Previous View.
     */
	public static int getActionBtn_Previous() { return EditorInfo.IME_ACTION_PREVIOUS;} // 7
	/**
     * General Text Data.
     */
	public static int getDataType_Text() { return EditorInfo.TYPE_CLASS_TEXT;}
	/**
     * Numbers Only Data.
     */
	public static int getDataType_Num() { return EditorInfo.TYPE_CLASS_NUMBER;}
	/**
     * Floating Point Data with Sign and Decimal Point.
     */
	public static int getDataType_Float() { return EditorInfo.TYPE_CLASS_NUMBER | EditorInfo.TYPE_NUMBER_FLAG_DECIMAL | EditorInfo.TYPE_NUMBER_FLAG_SIGNED;}
	/**
     * DateTime Data.
     */
	public static int getDataType_DateTime() { return EditorInfo.TYPE_CLASS_DATETIME;}
	/**
     * Date Data.
     */
	public static int getDataType_Date() { return EditorInfo.TYPE_CLASS_DATETIME | EditorInfo.TYPE_DATETIME_VARIATION_DATE;}
	/**
     * Time Data.
     */
	public static int getDataType_Time() { return EditorInfo.TYPE_CLASS_DATETIME | EditorInfo.TYPE_DATETIME_VARIATION_TIME;}
	/**
     * Password Data.
     */
	public static int getDataType_Password() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_PASSWORD;}
	/**
     * Email Address Data.
     */
	public static int getDataType_Email() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_EMAIL_ADDRESS;}
	/**
     * URL Text Data.
     */
	public static int getDataType_URL() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_URI;}
	/**
     * Phone Number Data.
     */
	public static int getDataType_Phone() { return EditorInfo.TYPE_CLASS_PHONE;}
	/**
     * Name Data.
     */
	public static int getDataType_Name() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_VARIATION_PERSON_NAME | EditorInfo.TYPE_TEXT_FLAG_CAP_WORDS;}
	/**
     * All Characters Uppercase.
     */
	public static int getDataType_Uppercase_All() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_FLAG_CAP_CHARACTERS;}
	/**
     * First Character in each Sentence Uppercase.
     */
	public static int getDataType_Uppercase_Sentences() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_FLAG_CAP_SENTENCES;}
	/**
     * First Character in each Word Uppercase.
     */
	public static int getDataType_Uppercase_Words() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_FLAG_CAP_WORDS;}
	/**
     * Text should have AutoCorrect Applied.
     */
	public static int getDataType_AutoCorrect() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_FLAG_AUTO_CORRECT | EditorInfo.TYPE_TEXT_FLAG_AUTO_COMPLETE;}
	/**
     * No Suggestions/AutoCorrect.
     */
	public static int getDataType_No_Suggestions() { return EditorInfo.TYPE_CLASS_TEXT | EditorInfo.TYPE_TEXT_FLAG_NO_SUGGESTIONS;}
	/**
     * "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz".
     */
	public static String getChar_Alpha() { return "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";}
	/**
     * "ABCDEFGHIJKLMNOPQRSTUVWXYZ".
     */
	public static String getChar_UpperAlpha() { return "ABCDEFGHIJKLMNOPQRSTUVWXYZ";}
	/**
     * "0123456789".
     */
	public static String getChar_Num() { return "0123456789";}
	/**
     * Sets Padding to give a View added with AddView() functions.
     */
	public void SetPadding(int Left, int Top, int Right, int Bottom) {
		ViewLeftPadding= Left;
		ViewRightPadding= Right;
		ViewTopPadding= Top;
		ViewBottomPadding= Bottom;
	}
	
	/**
     * Sets xAlignment for a View added with AddView() functions.
     */
	public void SetXalignment(int xAlign, String xView) {
		this.xAlign= xAlign;
		this.xView= xView;
	}
	
	/**
     * Sets yAlignment for a View added with AddView() functions.
     */
	public void SetYalignment(int yAlign, String yView) {
		this.yAlign= yAlign;
		this.yView= yView;
	}
	
	/**
     * Sets Width and Height for a View added with AddView() functions.
     */
	public void SetSize(int Width, int Height) {
		this.Width= Width;
		this.Height= Height;
	}
	
	/**
     * android.R.style.Theme_Black (Android 1.0+/API1)
     */
	public static int getTheme_Black() { return 16973832;}
	
	/**
     * android.R.style.Theme_DeviceDefault (Android 4.0+/API14)
     */
	public static int getTheme_DeviceDefault() {if (android.os.Build.VERSION.SDK_INT > 13) {return 16974120;}else {return 16973836;}}
	
	/**
     * android.R.style.Theme_Holo (Android 3.0+/API11)
     */
	public static int getTheme_Holo() {if (android.os.Build.VERSION.SDK_INT > 10) {return 16973931;}else {return 16973836;}}
	
	/**
     * android.R.style.Theme_Holo_Light (Android 3.0+/API11)
     */
	public static int getTheme_HoloLight() {if (android.os.Build.VERSION.SDK_INT > 10) {return 16973934;}else {return 16973836;}}
	
	/**
     * android.R.style.Theme_Light (Android 1.0+/API1)
     */
	public static int getTheme_Light() { return 16973836;}
	
	/**
     * android.R.style.Theme_Translucent (Android 1.0+/API1)
     */
	public static int getTheme_Translucent() { return 16973839;}
	
	private LinkedHashMap<String, View> viewList= new LinkedHashMap<String, View>();
	private RelativeLayout myRelLayout;
	private int Width=0, Height=0, xAlign=0, yAlign=0, ViewLeftPadding=0, ViewRightPadding=0, ViewTopPadding=0, ViewBottomPadding=0;
	private String xView="", yView="";

    @Override
    @Hide
    public void innerInitialize(BA pBA, String pEventName, boolean pKeepOldObject) {
        if (!pKeepOldObject) {
            setObject(new ScrollView(pBA.context));
        }
        super.innerInitialize(pBA, pEventName, true);
    }
    
    @Override
	public void Initialize(BA ba, String EventPrefix) {
		super.Initialize(ba, EventPrefix);
	}
    
    @Override
    @Hide
	public void RemoveView() {
		super.RemoveView();
	}

    /**
     * Sets the Relative Layout's Background Image.
     */
    @Override
	public void SetBackgroundImage(Bitmap Bitmap) {
		super.SetBackgroundImage(Bitmap);
	}

    /**
     * Changes the Relative Layout's Position and Size.
     */
    @Override
	public void SetLayout(int Left, int Top, int Width, int Height) {
		super.SetLayout(Left, Top, Width, Height);
	}

	@Override
	public void Invalidate2(Rect Rect) {
		super.Invalidate2(Rect);
	}

	@Override
	public void Invalidate3(int Left, int Top, int Right, int Bottom) {
		super.Invalidate3(Left, Top, Right, Bottom);
	}

	/**
     * Hides the Relative Layout View.
     */
    public void HideLayout(){
        getObject().setVisibility(View.GONE);
    }

    /**
     * Shows the Relative Layout View after it was hidden.
     */
    public void ShowLayout(){
        getObject().setVisibility(View.VISIBLE);
    }
    
    /**
     * Returns the Named View if it Exists or Null.
     */
    public View GetView(String viewName){
    	return viewList.get(viewName);
    }
    
    /**
     * Returns the Named Views in a Map.
     */
    public Map GetViews(){
    	Map newMap = new Map();
    	MyMap newViewMap = new MyMap();
    	newViewMap.putAll(viewList);
    	newMap.setObject(newViewMap);
    	return newMap;
    }
    
    /**
     * Hides the Specified View.
     */
    public void HideView(String viewName){
    	if (viewList.containsKey(viewName)) viewList.get(viewName).setVisibility(View.GONE);
    }

    /**
     * Shows the Specified View after it was hidden.
     */
    public void ShowView(String viewName){
    	if (viewList.containsKey(viewName)) viewList.get(viewName).setVisibility(View.VISIBLE);
    }
    
    /**
     * Enables a View optionally passing a Background (Pass 0 for No Background Change).
     * - Set Background to a 9 Patch or other Image with GetBackgroundResourceID
     */
    public void EnableView(String viewName, int BackgroundResourceID){
    	if (viewList.containsKey(viewName)){
    		viewList.get(viewName).setEnabled(true);
    		viewList.get(viewName).setFocusable(true);
    		viewList.get(viewName).setFocusableInTouchMode(true);
    		if (BackgroundResourceID != 0) viewList.get(viewName).setBackgroundResource(BackgroundResourceID);
    	}
    }

    /**
     * Disables a View optionally passing a Background (Pass 0 for No Background Change).
     * - Set Background to a 9 Patch or other Image with GetBackgroundResourceID
     */
    public void DisableView(String viewName, int BackgroundResourceID){
        if (viewList.containsKey(viewName)){
    		viewList.get(viewName).setEnabled(false);
    		viewList.get(viewName).setFocusable(false);
    		viewList.get(viewName).setFocusableInTouchMode(false);
    	}
        if (BackgroundResourceID != 0) viewList.get(viewName).setBackgroundResource(BackgroundResourceID);
    }
    
    /**
     * Get the ID of the given Image/9 Patch Resource Name.
     */
    public int GetBackgroundResourceID(String ImageName) {
    	return BA.applicationContext.getResources().getIdentifier(ImageName, "drawable", BA.packageName);
    }
    
    /**
     * Set the Theme of the Activity to one of the Theme_ Constants.
     */
    public void SetTheme(int ThemeID) {
    	ba.activity.setTheme(ThemeID);
    }

    /**
     * Scrolls to Top of Layout
     */
    public void ScrollTop() {
	   getObject().fullScroll(View.FOCUS_UP);
   }
    
    /**
     * Scrolls to Bottom of Layout
     */
    public void ScrollBottom() {
	   getObject().fullScroll(View.FOCUS_DOWN);
   }
    
    /**
     * Scrolls Up One Page in Layout
     */
    public void ScrollPgUp() {
	   getObject().pageScroll(View.FOCUS_UP);
   }
    
    /**
     * Scrolls Down One Page in Layout
     */
    public void ScrollPgDn() {
	   getObject().pageScroll(View.FOCUS_DOWN);
   }
    
    /**
     * Returns Current Selection # of Combobox or -1 for no selection/error.
     */
    public int GetCurrentSelection(String ViewName) {
    	if (viewList.containsKey(ViewName)){
    		View newView= viewList.get(ViewName);
    		if (newView instanceof Spinner) {
    			return (((Spinner) newView).getSelectedItemPosition());
    		}
    	}
    	return -1;
   }
    
    /**
     * Sets Current Selection # of Combobox.
     */
    public void SetCurrentSelection(String ViewName, int selIndex) {
    	if (viewList.containsKey(ViewName)){
    		View newView= viewList.get(ViewName);
    		if (newView instanceof Spinner) ((Spinner) newView).setSelection(selIndex);
    	}
   }
    
    /**
     * Returns Text Value of the View.
     */
    public String GetValue(String ViewName) {
    	if (viewList.containsKey(ViewName)){
    		View newView= viewList.get(ViewName);
    		if (newView instanceof Spinner) {
    			return (((Spinner) newView).getSelectedItem().toString());
    		}
    		else if (newView instanceof TextView) {
    			return (((TextView) newView).getText().toString());
			}
    	}
    	return "";
   }
    
    /**
     * Sets Text Value of the View.
     */
    public void SetValue(String ViewName, String Value) {
    	if (viewList.containsKey(ViewName)){
    		View newView= viewList.get(ViewName);
    		if (newView instanceof Spinner) {
    			for (int i = 0; i < (((Spinner) newView).getCount()); i++) {
    				if ((((Spinner) newView).getItemAtPosition(i).toString()) == Value){
    					((Spinner) newView).setSelection(i);
    					break;
    				}
				}
    		}
    		else if (newView instanceof TextView) {
    			((TextView) newView).setText(Value);
			}
    	}
   }
    
    /**
     * Returns a Map with the Current Text of all Named Views.
     */
    public Map SaveState() {
    	Map tempMap = new Map();
    	tempMap.Initialize();
    	for (java.util.Map.Entry<String, View> entry : viewList.entrySet()) {
    		if (entry.getValue() instanceof Spinner) {
    			tempMap.Put(entry.getKey(), (((Spinner) entry.getValue()).getSelectedItem().toString()));
    		}
    		else if (entry.getValue() instanceof TextView) {
    			tempMap.Put(entry.getKey(), (((TextView) entry.getValue()).getText().toString()));
			}
		}
    	return tempMap;
   }
    
    /**
     * Sets the text of the Named Views matching the Key Names in the Map to the Text Values in the Map.
     */
    public void RestoreState(Map Values) {
    	if (Values.IsInitialized()) {
    		for (int i = 0; i < Values.getSize(); i++) {
    			SetValue(Values.GetKeyAt(i).toString(), Values.GetValueAt(i).toString());
			}
    	}
   }
    
    /**
     * Sets the text of the Named Views matching the Key Names in the Cursor to the Text Values in the Map.
     */
    public void RestoreStateDB(Cursor Values) {
    	if (Values.getCount() == 1) {
    		Values.moveToFirst();
    		for (int i = 0; i < Values.getColumnCount(); i++) {
    			SetValue(Values.getColumnName(i), Values.getString(i));
			}
    	}
   }
    
    /**
     * Scrolls to Specific Location in Layout (Instant or Smooth)
     */
    public void ScrollToY(int Y, boolean Smooth) {
	   if (Smooth) {
		   getObject().smoothScrollTo(0, Y);
	   }
	   else {
		   getObject().scrollTo(0, Y);
	   }
   }
    
    /**
     * Adds a Label to the Relative Layout.
     * - Height and Width can be an actual value/DIP Value or sizeWrapContent/sizeMatchParent.
     * - If your xAlign/yAlign rule is set to one of the rules in relation to another view then pass it's Name in as the xView/yView otherwise set them to "" for Parent Relationships.
     * - If MaxLines > 0 then the Label will only wrap that many lines and Marquee shift the overflow if set to 1.
     * - LinkedView is the ViewName of the view you want given focus when the label is clicked.
     * - Set Background to a 9 Patch or other Image using BackgroundResourceID with GetBackgroundResourceID
     */
    public LabelWrapper AddLabel(String ViewName, String EventPrefix, String Text, float TextSize, Boolean Bold, int MaxLines, int TextColor, String LinkedView, int BackgroundResourceID){
    	LabelWrapper b4aLabel = new LabelWrapper();
    	b4aLabel.Initialize(ba, EventPrefix);
    	TextView NewView = b4aLabel.getObject();
    	
    	if (BackgroundResourceID != 0) NewView.setBackgroundResource(BackgroundResourceID);
    	
    	NewView.setText(Text);
    	NewView.setTextSize(TextSize);
    	if (Bold) NewView.setTypeface(Typeface.DEFAULT_BOLD);
    	NewView.setTextColor(TextColor);
    	if (Width < 0) NewView.setMinEms(1);
    	if (MaxLines > 0) {
    		NewView.setMinLines(1);
    		NewView.setMaxLines(MaxLines);
    		if (MaxLines == 1) {
    			NewView.setSingleLine(true);
    			NewView.setHorizontallyScrolling(true);
        		NewView.setEllipsize(TextUtils.TruncateAt.MARQUEE);
        		NewView.setMarqueeRepeatLimit(-1);
        		NewView.setSelected(true);
    		}
    		else {
    			NewView.setSingleLine(false);
			}
    	}
    	if (LinkedView.length() > 0) {
    		NewView.setTag(LinkedView);
    		NewView.setOnClickListener(new OnClickListener() {
				@Override
				public void onClick(View myView) {
					if (viewList.containsKey(myView.getTag())) viewList.get(myView.getTag()).requestFocus();
				}
			});
    	}
    	AddView(NewView, ViewName);
    	return b4aLabel;
    }
    
    /**
     * Adds a Combobox/Spinner to the Relative Layout.
     * - Height and Width can be an actual value/DIP Value or sizeWrapContent/sizeMatchParent.
     * - If your xAlign/yAlign rule is set to one of the rules in relation to another view then pass it's Name in as the xView/yView otherwise set them to "" for Parent Relationships.
     * - Set Background to a 9 Patch or other Image using BackgroundResourceID with GetBackgroundResourceID
     */
	public SpinnerWrapper AddCombobox(String ViewName, String EventPrefix, String Prompt, float TextSize, int TextColor, List Items, int BackgroundResourceID){
    	SpinnerWrapper b4aSpinner = new SpinnerWrapper();
    	b4aSpinner.Initialize(ba, EventPrefix);
    	Spinner NewView = b4aSpinner.getObject();
    	if (BackgroundResourceID != 0) NewView.setBackgroundResource(BackgroundResourceID);
    	b4aSpinner.setPrompt(Prompt);
    	b4aSpinner.setTextSize(TextSize);
    	b4aSpinner.setTextColor(TextColor);
    	if (Items.IsInitialized()) b4aSpinner.AddAll(Items);

    	AddView(NewView, ViewName);
    	return b4aSpinner;
    }
    
    /**
     * Adds a Button to the Relative Layout.
     * - Height and Width can be an actual value/DIP Value or sizeWrapContent/sizeMatchParent.
     * - If your xAlign/yAlign rule is set to one of the rules in relation to another view then pass it's Name in as the xView/yView otherwise set them to "" for Parent Relationships.
     * - If MaxLines > 0 then the Button will only wrap that many lines and Marquee shift the overflow if set to 1.
     * - Set Background to a 9 Patch or other Image using BackgroundResourceID with GetBackgroundResourceID
     */
    public ButtonWrapper AddButton(String ViewName, String EventPrefix, String Text, float TextSize, Boolean Bold, int MaxLines, int TextColor, final int UpBackgroundResourceID, final int DownBackgroundResourceID){
    	ButtonWrapper b4aButton = new ButtonWrapper();
    	b4aButton.Initialize(ba, EventPrefix);
    	Button NewView = b4aButton.getObject();
    	if (UpBackgroundResourceID != 0) NewView.setBackgroundResource(UpBackgroundResourceID);
    	
    	if ((UpBackgroundResourceID != 0) && (DownBackgroundResourceID != 0)) {
    		NewView.setOnTouchListener(new OnTouchListener() {
				
				public boolean onTouch(View MyView, MotionEvent Event) {
					switch (Event.getAction()) {
					case MotionEvent.ACTION_DOWN:
						MyView.setBackgroundResource(DownBackgroundResourceID);
						break;
					case MotionEvent.ACTION_UP:
						MyView.setBackgroundResource(UpBackgroundResourceID);
						break;
					case MotionEvent.ACTION_MOVE:
						MyView.getParent().requestDisallowInterceptTouchEvent(true);
						break;
					}
					return false;
				}
			});
    	}
    	
    	NewView.setText(Text);
    	NewView.setTextSize(TextSize);
    	if (Bold) NewView.setTypeface(Typeface.DEFAULT_BOLD);
    	NewView.setTextColor(TextColor);
    	if (Width < 0) NewView.setMinEms(1);
    	if (MaxLines > 0) {
    		NewView.setMinLines(1);
    		NewView.setMaxLines(MaxLines);
    	}
    	AddView(NewView, ViewName);
    	return b4aButton;
    }
    
    /**
     * Adds a Textbox to the Relative Layout.
     * - Height and Width can be an actual value/DIP Value or sizeWrapContent/sizeMatchParent.
     * - If your xAlign/yAlign rule is set to one of the rules in relation to another view then pass it's Name in as the xView/yView otherwise set them to "" for Parent Relationships.
     * - DateType is one of the DataType_Type values to indicate the type of Data to be input.
     * - CharFilter is a string of chars to limit input to. There are some Char_Value strings to make things easier.
     * - ActionBtn Determines what Button appears at the bottom right of keyboard. Fires off ViewEventPrefix_imeAction event when pressed.
     * - MinChars set smallest size in chars of Textbox.
     * - MaxChars sets max characters allowed in the field. Also sets Max EMS size to aid in dynamic content sizing.
     * - Lines of 0 does nothing, 1 forces single line, > 1 sets multi-line and limits to that many lines tall visible if you set Height to 0.
     * - AutoCompleteItems is a list of items to show in a dropdown as you type that match what has been typed so far.
     * - Set Background to a 9 Patch or other Image using BackgroundResourceID with GetBackgroundResourceID
     */
    public AutoCompleteEditTextWrapper AddTextbox(String ViewName, final String EventPrefix, int DataType, final String CharFilter, int ActionBtn, int MinChars, int MaxChars, int Lines, String Text, float TextSize, int TextColor, String Hint, int HintColor, List AutoCompleteItems, int BackgroundResourceID, final String NextView, final String PrevView){
    	AutoCompleteEditTextWrapper b4aTextbox = new AutoCompleteEditTextWrapper();
    	b4aTextbox.Initialize(ba, EventPrefix);
    	EditText NewView = b4aTextbox.getObject();
    	
    	if (BackgroundResourceID != 0) NewView.setBackgroundResource(BackgroundResourceID);
    	
    	NewView.setText(Text);
    	NewView.setTextSize(TextSize);
    	NewView.setTypeface(Typeface.DEFAULT);
    	NewView.setTextColor(TextColor);
    	NewView.setHint(Hint);
    	NewView.setHintTextColor(HintColor);
    	NewView.setMinEms(MinChars);
    	NewView.setScrollbarFadingEnabled(true);
    	if ((MaxChars > 0) || ((DataType & EditorInfo.TYPE_TEXT_FLAG_CAP_CHARACTERS) > 0)) {
	    	if (MaxChars > 0) NewView.setMaxEms(MaxChars);
	    	if ((MaxChars > 0) && ((DataType & EditorInfo.TYPE_TEXT_FLAG_CAP_CHARACTERS) > 0)) {
	    		InputFilter[] FilterArray = new InputFilter[2];
	    		FilterArray[0] = new InputFilter.LengthFilter(MaxChars);
		        FilterArray[1] = new InputFilter.AllCaps();
		        NewView.setFilters(FilterArray);
	    	}
	    	else {
	    		InputFilter[] FilterArray = new InputFilter[1];
	    		if (MaxChars > 0) {
	    			FilterArray[0] = new InputFilter.LengthFilter(MaxChars);
	    		}
	    		else {
	    			FilterArray[0] = new InputFilter.AllCaps();
				}		
		        NewView.setFilters(FilterArray);
			}
    	}
    	if (Lines > 0) {
    		if (Lines > 1) {
    			NewView.setSingleLine(false);
    			NewView.setGravity(Gravity.TOP | Gravity.LEFT);
    			NewView.setInputType(DataType | InputType.TYPE_TEXT_FLAG_MULTI_LINE);
    			NewView.setImeOptions(EditorInfo.IME_FLAG_NO_ENTER_ACTION | ActionBtn);
    			if (Height == 0) Height = (NewView.getLineHeight() * Lines) + NewView.getPaddingTop() + NewView.getPaddingBottom();
    			NewView.setOnTouchListener(new OnTouchListener() {
					
					public boolean onTouch(View MyView, MotionEvent Event) {
						if (Event.getAction() == MotionEvent.ACTION_MOVE) {
							MyView.getParent().requestDisallowInterceptTouchEvent(true);
						}
						return false;
					}
				});
    		}
    		else {
        		NewView.setSingleLine(true);
    			NewView.setInputType(DataType);
    			NewView.setImeOptions(EditorInfo.IME_FLAG_NO_EXTRACT_UI | ActionBtn);
    		}
    	}
    	else {
    		NewView.setInputType(DataType | InputType.TYPE_TEXT_FLAG_IME_MULTI_LINE);
			NewView.setImeOptions(ActionBtn);
		}
    	
    	if (CharFilter.length() > 0) {
    		final int InputType = NewView.getInputType();
    		NewView.setKeyListener(new NumberKeyListener() {

                @Override
                protected char[] getAcceptedChars() {
                    return CharFilter.toCharArray();
                }

                @Override
                public int getInputType() {
                    return InputType;
                }
                
            });
    	}
		NewView.setOnEditorActionListener(new TextView.OnEditorActionListener() {

            @Override
            public boolean onEditorAction(TextView v, int actionId, KeyEvent event) {
            	if (actionId == EditorInfo.IME_ACTION_NEXT) {
            		if (NextView.length() > 0) {
            			if (viewList.containsKey(NextView)) return viewList.get(NextView).requestFocus();
            		}
            		return false;
            	}
            	if (actionId == EditorInfo.IME_ACTION_PREVIOUS) {
            		if (PrevView.length() > 0) {
            			if (viewList.containsKey(PrevView)) return viewList.get(PrevView).requestFocus();
            		}
            		return false;
            	}
            	if (EventPrefix.length() > 0) {
	            	Boolean b =  (Boolean)ba.raiseEvent(v, EventPrefix + "_imeAction");
	                if (b != null && b == true)
	                    return true;
	                else
	                    return false;
            	}
            	else {
					return false;
				}
            }                
        });
    	if (AutoCompleteItems.IsInitialized()) b4aTextbox.SetItems(ba, AutoCompleteItems);
    	AddView(NewView, ViewName);
    	return b4aTextbox;
    }
        
    /**
     * Adds a View to the Relative Layout.
     * - Height and Width can be an actual value/DIP Value or sizeWrapContent/sizeMatchParent.
     * - If your xAlign/yAlign rule is set to one of the rules in relation to another view then pass it's Name in as the xView/yView otherwise set them to "" for Parent Relationships.
     */
    public void AddView(View NewView, String ViewName){
    	if (myRelLayout == null) {
    		myRelLayout= new RelativeLayout(ba.context);
    		LayoutParams myLayoutParams = new LayoutParams(LayoutParams.MATCH_PARENT, LayoutParams.MATCH_PARENT);
            myRelLayout.setLayoutParams(myLayoutParams);
        	getObject().addView(myRelLayout);
    	}
    	RelativeLayout.LayoutParams myLayoutParams = new RelativeLayout.LayoutParams(Width, Height);
    	myLayoutParams.setMargins(ViewLeftPadding, ViewTopPadding, ViewRightPadding, ViewBottomPadding);
    	if (xView.length() == 0) { // Assume Parent Align
    		if (xAlign == getxAlign_ParentLeft()) myLayoutParams.addRule(getxAlign_ParentLeft());
    		if (xAlign == getxAlign_ParentRight()) myLayoutParams.addRule(getxAlign_ParentRight());
    		if (xAlign == getxAlign_ParentCentHorz()) myLayoutParams.addRule(getxAlign_ParentCentHorz());
    	}
    	else {
    		if (viewList.containsKey(xView)){
    			if (xAlign == getxAlign_ViewLeft()) myLayoutParams.addRule(getxAlign_ViewLeft(), viewList.get(xView).getId());
        		if (xAlign == getxAlign_ViewRight()) myLayoutParams.addRule(getxAlign_ViewRight(), viewList.get(xView).getId());
        		if (xAlign == getxAlign_LeftOfView()) myLayoutParams.addRule(getxAlign_LeftOfView(), viewList.get(xView).getId());
        		if (xAlign == getxAlign_RightOfView()) myLayoutParams.addRule(getxAlign_RightOfView(), viewList.get(xView).getId());
    		}
		}
    	if (yView.length() == 0) { // Assume Parent Align
    		if (yAlign == getyAlign_ParentTop()) myLayoutParams.addRule(getyAlign_ParentTop());
    		if (yAlign == getyAlign_ParentCentVert()) myLayoutParams.addRule(getyAlign_ParentCentVert());
    	}
    	else {
    		if (viewList.containsKey(yView)){
    			if (yAlign == getyAlign_AboveView()) myLayoutParams.addRule(getyAlign_AboveView(), viewList.get(yView).getId());
        		if (yAlign == getyAlign_BelowView()) myLayoutParams.addRule(getyAlign_BelowView(), viewList.get(yView).getId());
        		if (yAlign == getyAlign_ViewTop()) myLayoutParams.addRule(getyAlign_ViewTop(), viewList.get(yView).getId());
        		if (yAlign == getyAlign_ViewBottom()) myLayoutParams.addRule(getyAlign_ViewBottom(), viewList.get(yView).getId());
        		if (yAlign == getyAlign_ViewBaseline()) myLayoutParams.addRule(getyAlign_ViewBaseline(), viewList.get(yView).getId());
    		}
		}
    	NewView.setLayoutParams(myLayoutParams);
    	myRelLayout.addView(NewView);
    	if (ViewName.length() > 0) viewList.put(ViewName, NewView);
    }
    
    /**
     * Enables handling the Keyboard HeightChanged event. This event is raised when the soft keyboard size changes.
     * - Note that this event will not be raised in full screen activities (an Android limitation).
     */
    public void HandleKeyboardHeightChange(BA ba, String EventPrefix) {
        if (ba.vg.getParent() instanceof BALayout)
            return;
        ExtendedBALayout e = new ExtendedBALayout(ba.context, EventPrefix, ba);
        ba.activity.setContentView(e);
        BALayout.LayoutParams lp = new BALayout.LayoutParams();
        lp.height = ba.vg.getLayoutParams().height;
        lp.width = ba.vg.getLayoutParams().width;
        ba.vg.setLayoutParams(lp);
        e.addView(ba.vg);
        ba.activity.getWindow().setSoftInputMode(ba.activity.getWindow().getAttributes().softInputMode | WindowManager.LayoutParams.SOFT_INPUT_ADJUST_RESIZE);
    }
    
    static class ExtendedBALayout extends BALayout {
        private int lastHeight = -1;
        private final String eventName;
        private final BA ba;
        public ExtendedBALayout(Context context, String EventName, BA ba) {
            super(context);
            this.eventName = EventName.toLowerCase(BA.cul);
            this.ba = ba;
        }
        @Override
        protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
            int height = MeasureSpec.getSize(heightMeasureSpec);
            if (height != lastHeight && lastHeight != -1) {
                ba.raiseEvent(null, eventName + "_imeHeightChanged", height, lastHeight);
            }
            lastHeight = height;
            super.onMeasure(widthMeasureSpec, heightMeasureSpec);       
        }

    }
}